import os
from distutils import dir_util

from ._internal import defaultarguments, DoreahConfig

config = DoreahConfig("packageutils",
	packagename=None,
	populate=None
)


#if config["packagename"] is None: config(packagename=)

try:
	HOME_DIR = os.environ["XDG_DATA_HOME"].split(":")[0]
	assert os.path.exists(HOME_DIR)
except Exception:
	HOME_DIR = os.path.join(os.environ["HOME"],".local/share/")

DATA_DIR = None # the actual folder for this package

def _set_datafolder():
	global DATA_DIR
	if DATA_DIR is None:
		DATA_DIR = os.path.join(HOME_DIR,config["packagename"])
		os.makedirs(DATA_DIR,exist_ok=True)

def pkgdata(*args):
	_set_datafolder()
	return os.path.join(DATA_DIR,*args)

def init_datafolder():
	_set_datafolder()
	dir_util.copy_tree(config["populate"],DATA_DIR,update=False)



"""Module that offers data structures that are kept persistent on the file system"""

import os, shutil
import yaml, json, configparser
import re, ast


class DiskDict(dict):
	"""This object acts as a dictionary with persistence. It is not meant for big data,
	but simple configuration files and such. Keys are always strings and case-insensitive."""

	def __init__(self,filename,readonly=False,discard_none=False):
		self.filename = filename
		self.internal_dict = {}
		self.handler = get_handler(filename)
		self.readonly = readonly
		self.discard_none = discard_none

		if os.path.exists(self.filename):
			self._sync_from_disk()

		if self.discard_none:
			self.internal_dict = {k:v for k,v in self.internal_dict.items() if v is not None}


		try:
			self._sync_to_disk()
		except PermissionError as e:
			self.readonly = True

	def _sync_from_disk(self):
		data = self.handler.load_file(self.filename)
		data = {k.lower():data[k] for k in data}
		self.internal_dict = data
		#self.internal_dict.update(data)
	def _sync_to_disk(self):
		if not self.readonly:
			tmpfile = self.filename + ".tmp"
			data = {k.lower():self.internal_dict[k] for k in self.internal_dict}
			self.handler.write_file(tmpfile,data)
			shutil.move(tmpfile,self.filename)

	def __repr__(self):
		return self.internal_dict.__repr__()

	def __setitem__(self,key,value):
		if self.readonly: raise PermissionError
		self._sync_from_disk()
		self.internal_dict.__setitem__(key.lower(),value)
		self._sync_to_disk()

	def __getitem__(self,key):
		self._sync_from_disk()
		return self.internal_dict.__getitem__(key.lower())

	def __delitem__(self,key):
		if self.readonly: raise PermissionError
		self._sync_from_disk()
		self.internal_dict.__delitem__(key.lower())
		self._sync_to_disk()

	def get(self,key):
		return self.internal_dict.get(key)

	# update should save to disk only once. no need to do it for every entry
	def update(self,otherdict):
		if self.readonly: raise PermissionError
		self._sync_from_disk()
		self.internal_dict.update(otherdict)
		self._sync_to_disk()




# Handlers for the different data file types.

handlers = {}

def get_handler(filename):
	ext = filename.split('.')[-1].lower()
	return handlers[ext]


class GenericHandler:
	def __init_subclass__(cls):
		for ext in cls.extensions:
			handlers[ext] = cls()

	def load_file(self,filename):
		with open(filename) as descriptor:
			return self.load_descriptor(descriptor)

	def write_file(self,filename,data):
		with open(filename,"w") as descriptor:
			return self.write_descriptor(descriptor,data)


	def load_descriptor(self,descriptor):
		txt = descriptor.read()
		return self.text_to_data(txt)

	def write_descriptor(self,descriptor,data):
		txt = self.data_to_text(data)
		return descriptor.write(txt)


class YAMLHandler(GenericHandler):
	extensions = ["yml","yaml"]

	def load_descriptor(self,descriptor):
		return yaml.safe_load(descriptor)

	def write_descriptor(self,descriptor,data):
		return yaml.dump(data,descriptor)

class JSONHandler(GenericHandler):
	extensions = ["json"]

	def load_descriptor(self,descriptor):
		return json.load(descriptor)

class INIHandler(GenericHandler):
	extensions = ["ini"]

	special_identifiers = [
		(True,['true','yes','y']),
		(False,['false','no','n']),
		(None,['none','null'])
	]


	def parse_dict_values(self,inputdict):
		outputdict = {}
		for key,val in inputdict.items():
			for resolve_to,values in self.special_identifiers:
				if val.lower() in values:
					val = resolve_to
					break
			else:
				try:
					val = ast.literal_eval(val)
				except Exception:
					val = val # just use as string
			outputdict[key] = val
		return outputdict

	def load_file(self,filename):
		try:
			cfg = configparser.ConfigParser(interpolation=None)
			cfg.read(filename)
			try:
				return self.parse_dict_values(cfg['MALOJA'])
			except KeyError:
				return {}
		except configparser.MissingSectionHeaderError:
			with open(filename,'r') as fd:
				return self.text_to_data("[MALOJA]\n\n" + fd.read())

	def text_to_data(self,txt):
		cfg = configparser.ConfigParser(interpolation=None)
		cfg.read_string(txt)
		return self.parse_dict_values(cfg['MALOJA'])

	def write_descriptor(self,descriptor,data):
		cfg = configparser.ConfigParser(interpolation=None)
		cfg['MALOJA'] = data
		cfg.write(descriptor)



from bs4 import BeautifulSoup
from copy import deepcopy
import re
import os
import hashlib
import copy

from ._internal import DEFAULT, defaultarguments, DoreahConfig
from .datatypes import DictStack


config = DoreahConfig("pyhp",
	interpret=str,
	precompile=True
)



precompiled = {}



def compile(raw):

	soup = BeautifulSoup(raw,"html.parser")

	compile_node_expressions(soup)
	return soup

def compile_node_expressions(node):
	if isinstance(node,str): return

	if node.name == "pyhp":
		for a in ["in","with","save","if"]:
			if _attr(node,a) is not None:
				node[a] = compile(_attr(node,a))

	for subnode in node:
		compile_node_expressions(subnode)



@defaultarguments(config,interpret="interpret",precompile="precompile")
def file(path,d={},interpret=DEFAULT,noroot=False,precompile=DEFAULT):
	"""Parses a pyhp source file and returns the generated html code.

	:param string path: Location of the pyhp source file
	:param dict d: Variables dictionary
	:param function interpret: Function that determines string representations of variables
	:param bool precompile: Whether compiled pyhp pages should be cached in memory for faster page generation
	:return: HTML source
	"""

	filepath = os.path.abspath(path)
	directory = os.path.dirname(filepath)

	with open(path,"r") as f:
		content = f.read()

	if precompile:

		h = hashlib.md5()
		h.update(content.encode())
		check = h.digest()
		if filepath in precompiled and precompiled[filepath]["checksum"] == check:
			pass
		else:
			precompiled[filepath] = {"object":compile(content),"checksum":check}
		content = copy.copy(precompiled[filepath]["object"])



	return parse(content,d,interpret=interpret,directory=directory,noroot=noroot)


def _file(path,d,interpret=DEFAULT,noroot=False):

	with open(path,"r") as f:
		content = f.read()

	directory = os.path.dirname(os.path.abspath(path))

	return _parse(content,d,interpret=interpret,directory=directory,noroot=noroot)


@defaultarguments(config,interpret="interpret")
def parse(src,d={},interpret=DEFAULT,directory=None,noroot=False):
	"""Parses pyhp source and returns the generated html code.

	:param string src: Source string
	:param dict d: Variables dictionary
	:param function interpret: Function that determines string representations of variables
	:return: HTML source
	"""

	doc = _parse(src,d,interpret=interpret,directory=directory,noroot=noroot)


	#raw = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
	#raw = etree.tostring(doc,encoding="unicode",method="html").replace("&gt;",">").replace("&lt;","<")
	return doc.decode()

# to tree
def _parse(src,d,interpret=DEFAULT,directory=None,noroot=False):

	if not isinstance(d,DictStack): d = DictStack(d)

	if isinstance(src,str):
		doc = BeautifulSoup(src,"html.parser")
	elif isinstance(src,BeautifulSoup):
		doc = src
	else:
		raise Exception
	doc = _parse_node(doc,d,interpret,directory=directory)[0]
	#result = list(doc.children)

	if noroot:
		return list(doc.children)
	else:
		return doc


def _parse_node(node,d,interpret,directory=None):

	if isinstance(node,str): return [node]


	## parse pyhp nodes

	if node.name == "pyhp":

		### CODE

		if len(node.attrs) == 0:
			#print("Executing code!")
			#print({k:d[k] for k in d if not k.startswith("_")})
			code = list(node.children)[0]
			code = code.strip("\t").strip(" ")
			code = code.split("\n")
			if len(code) == 1:
				# one-line code
				code = code[0]
			else:
				# multiline code
				if code[0] != "" or code[-1] != "":
					print("Malformed code block!")
				code = code[1:-1]
				roottabs = 0
				for char in code[0]:
					if char == "\t" or char == " ":
						roottabs += 1
					else:
						break

				code = [line[roottabs:] for line in code]
				code = "\n".join(code)

			exec(code,{},d)

			return []

		#### SAVE

		if _attr(node,"save") is not None and _attr(node,"as") is not None:
			#print("d[" + _attr(node,"as") + "] = " + _attr(node,"save"))
			d[_attr(node,"as")] = eval(_attr(node,"save"),{},d)
			#print({k:d[k] for k in d if not k.startswith("_")})

			return []

		### IMPORT

		if _attr(node,"import") is not None:
			exec("import " + _attr(node,"import"),{},d)

			return []


		#### INCLUDE

		elif _attr(node,"include") is not None:

			filename = _attr(node,"include")
			if directory is None:
				# relative to execution path
				filename = filename
			else:
				# relative to this file
				filename = os.path.join(directory,filename)


			if _attr(node,"with") is not None:

				localdict = eval(_attr(node,"with"),{},d)

				d.push(localdict)
			#	hidedict = {}
			#	# save overridden variables
			#	for key in localdict:
			#		if key in d:
			#			hidedict[key] = d[key]
			#	d.update(localdict)




			try:
				subnodes = _file(filename,d,interpret=interpret,noroot=True)
				for attr in node.attrs:
					# give attributes to included first top node
					if attr not in ["with","include"]:
						subnodes[0].attrs[attr] = _attr(node,attr)

			except Exception:
				print("Could not include",filename)
				raise
				subnodes = []

			if _attr(node,"with") is not None:
				# restore outer environment
			#	for key in localdict:
			#		del d[key]
			#	for key in hidedict:
			#		d[key] = hidedict[key]
				d.pop()

			return subnodes


		#### IF

		elif _attr(node,"if") is not None:
			if eval(_attr(node,"if"),{},d):
				nodestoreturn = []
				for sn in node:
					nodestoreturn += _parse_node(sn,d,interpret,directory=directory)
				return nodestoreturn
			else:
				return []


		#### FOR IN

		elif _attr(node,"for") is not None and _attr(node,"in") is not None:
			nodestoreturn = []
			# for loop of the elements
			first = True
			try:
				elements = eval(_attr(node,"in"),{},d)
			except Exception:
				# allow invalid expressions in for loops, just ignore them
				elements = []
			for element in elements:
				if not first and _attr(node,"separator") is not None:
					nodestoreturn += [_attr(node,"separator")]
				first = False

				# in case we overload a dict entry, keep the old one
			#	sentinel = object()
			#	if _attr(node,"for") in d:
			#		hide = d[_attr(node,"for")]
			#	else:
			#		hide = sentinel
			#
			#
			#	# the dict needs to remain the same object so changes from one node
			#	# in the for loop are carried over into the next loop
			#	d.update({_attr(node,"for"):element})

				d.push({_attr(node,"for"):element})

				# now go through the nodes each time
				for sn in node:
					nodestoreturn += _parse_node(deepcopy(sn),d,interpret,directory=directory)

				# clear the variable after each loop
			#	del d[_attr(node,"for")]
			#	if hide is not sentinel:
			#		d[_attr(node,"for")] = hide
				d.pop()

			return nodestoreturn


		### LOCAL SCOPE

		elif _attr(node,"with") is not None:

			localdict = eval(_attr(node,"with"),{},d)
		#	hidedict = {}
		#
		#	# save overridden variables
		#	for key in localdict:
		#		if key in d:
		#			hidedict[key] = d[key]
		#
		#	d.update(localdict)
			d.push(localdict)

			nodestoreturn = []
			for sn in node:
				nodestoreturn += _parse_node(deepcopy(sn),d,interpret,directory=directory)

			# restore outer environment
		#	for key in localdict:
		#		del d[key]
		#	for key in hidedict:
		#		d[key] = hidedict[key]
			d.pop()

			return nodestoreturn


		#### ECHO

		elif _attr(node,"echo") is not None:
			return BeautifulSoup(interpret(eval(_attr(node,"echo"),{},d)),"html.parser")
			#return [interpret(eval(_attr(node,"echo"),d))]


		else:
			print("Not a valid pyhp tag!")
			return []


	## parse normal nodes

	else:

		## replace attributes (not necessary in pyhp nodes)

		if hasattr(node,"attrs"):
			for name,value in node.attrs.items():
				if isinstance(value,list): value = " ".join(value)
				vars = re.findall("{.*?}",value)
				for v in vars:
					vname = v[1:-1]
					try:
						value = value.replace(v,interpret(eval(vname,{},d)))
					except Exception:
						pass
						print("Error parsing:",v,"in attribute")

				node.attrs[name] = value



		subnodes = [n for n in node]
		newsubnodes = []
		for subnode in subnodes:
			sn = subnode.extract()
			newsubnodes += _parse_node(sn,d,interpret,directory=directory)

		for nsn in newsubnodes:
			node.append(nsn)


		return [node]



def _attr(node,name):
	return node.get(name)


### run test server

from .control import mainfunction

@mainfunction({"p":"port","h":"host"},shield=True)
def run_testserver(port=1337,host="::"):
	from bottle import get, run, static_file
	from doreah.pyhp import file
	import os

	@get("/<path:path>")
	def serve_file(path):

		if os.path.exists(path + ".pyhp"):
			return file(path + ".pyhp")
		if os.path.exists(path):
			return static_file(path,root="")
		if os.path.exists(path + ".html"):
			return static_file(path + ".html",root="")

		return static_file(path,root="") # will produce proper error


	run(host=host,port=port)



from ._internal import DEFAULT, defaultarguments, DoreahConfig

config = DoreahConfig("filesystem2",
	ignore_dotfiles=True
)


import os
import shutil
import math


# ABC

class AbstractPath:
	def parent(self):
		try:
			parenttype = self.IS_IN
		except Exception:
			parenttype = self.__class__
		return parenttype(self.path[:-1])


class AbstractDirectory(AbstractPath):
	def mount(self,pth):
		#return self.__class__(self.path + pth.path)
		res = combines[(type(self),type(pth))]
		return res(self.path + pth.path)

	def __add__(self, other):
		return self.mount(other)

	def join(self,pth):
		return os.path.join(self.__path__(),pth)

class AbstractFile(AbstractPath):
	pass



# ABSOLUTE

class AbsolutePath(AbstractPath):
	def __init__(self,path):
		if isinstance(path,str):
			path = os.path.abspath(path)
			path = path.strip("/")
			self.path = path.split("/")
		else:
			self.path = path

	def __path__(self):
		return os.path.join("/",*self.path)

	def exists(self):
		return os.path.exists(self.__path__())


class AbsoluteDirectory(AbstractDirectory,AbsolutePath):
	def __getitem__(self,node):
		trgt = self.join(node)
		if os.path.exists(trgt):
			return AbsoluteFile(trgt)
		else:
			os.makedirs(trgt,exist_ok=True)
			return AbsoluteDirectory(trgt)

	def _listdir(self):
		return os.listdir(self.__path__())

	def files(self):
		for e in self._listdir():
			if os.path.isfile(self.join(e)):
				yield RelativeFile(e)

	def folders(self):
		for e in self._listdir():
			if not os.path.isfile(self.join(e)):
				yield RelativeDirectory(e)

	def all_files_relative(self,maxdepth=math.inf):
		"""Returns all files inside this directory, relative to it"""
		if maxdepth == 0: return
		for f in self.files():
			yield f
		for f in self.folders():
			for ff in self.mount(f).all_files_relative(maxdepth=maxdepth-1):
				yield f.mount(ff)

	def copyto(self,trgt):
		"""Copies this directory to the target directory"""
		assert not os.path.exists(trgt.__path__())
		shutil.copytree(self.__path__(),trgt.__path__())

	def copypath(self,relpth,targetdir):
		"""Copies the given relative path from this directory to the target directory, preserving the full path"""
		src = self.mount(relpth)
		trgt = targetdir.mount(relpth)
		src.copyto(trgt)



class AbsoluteFile(AbstractFile,AbsolutePath):
	IS_IN = AbsoluteDirectory

	def copyto(self,trgt):
		assert not os.path.exists(trgt.__path__())
		shutil.copy(self.__path__(),trgt.__path__())

	def open(self,mode="r"):
		return open(self.__path__(),mode=mode)


# RELATIVE

class RelativePath(AbstractPath):
	def __init__(self,path):
		if isinstance(path,str):
			path = path.strip("/")
			self.path = path.split("/")
		else:
			self.path = path


	def mountin(self,pth):
		return pth.mount(self)

	def __path__(self):
		return os.path.join(*self.path)

class RelativeDirectory(AbstractDirectory,RelativePath):
	pass

class RelativeFile(AbstractFile,RelativePath):
	IS_IN = RelativeDirectory
	pass



combines = {
	(AbsoluteDirectory,RelativeDirectory): AbsoluteDirectory,
	(AbsoluteDirectory,RelativeFile): AbsoluteFile,
	(RelativeDirectory,RelativeDirectory): RelativeDirectory,
	(RelativeDirectory,RelativeFile): RelativeFile

}



import datetime
import inspect
import os
import shutil

from ._internal import defaultarguments, gopen, DoreahConfig
from ._color import _ANSICOLOR


_queue = []
_locked = False


# logfolder			folder to store logfiles in
# timeformat		strftime format for log files
# defaultmodule		name for the main running script
# verbosity			higher means more (less important) messages are shown on console
config = DoreahConfig("logging",
	logfolder="logs",
	timeformat="%Y/%m/%d %H:%M:%S",
	defaultmodule="main",
	only_leaf_module=True,
	verbosity=0,
	maxsize=4194304
)



def log(*entries,module=None,heading=None,indent=0,importance=0,color=None):
	"""Logs all supplied arguments, separate line for each. Only writes to logfile if importance value is lower than the set verbosity value.

	:param string entries: All log entries, one line per entry
	:param string module: Custom category. Log entry will be prepended by this string in the console and be written to this file on disk. Defaults to actual name of the python module.
	:param integer heading: Heading category. Headings will be visually distinguished from normal entries.
	:param integer indent: Indent to be added to entry
	:param integer importance: Low means important. If this value is higher than the verbosity, entry will not be shown on console.
	:param string color: Color-codes console output. Can be a hex string, a RGB tuple or the name of a color."""

	now = datetime.datetime.utcnow().strftime(config["timeformat"])


	colorprefix,colorsuffix = _ANSICOLOR(color)

	# log() can be used to add empty line
	if len(entries) == 0: entries = ("",)

	# make it easier to log data structures and such
	entries = tuple([str(msg) for msg in entries])

	# header formating
	if heading == 2:
		entries = ("","","####") + entries + ("####","")
	elif heading == 1:
		entries = ("","","","# # # # #","") + entries + ("","# # # # #","","")

	# indent
	prefix = "\t" * indent

	# module name
	if module is None:
		try:
			module = inspect.getmodule(inspect.stack()[1][0])
			module = getattr(module,"__logmodulename__",module.__name__)
			if config["only_leaf_module"]: module = module.split(".")[-1]
			if module == "__main__": module = config["defaultmodule"]
		except Exception:
			module = "interpreter"

	global _locked, _queue
	if _locked:
		for msg in entries:
			_queue.append({"time":now,"prefix":prefix,"msg":msg,"module":module,"console":(importance <= config["verbosity"])})
	else:
		# console output
		if (importance <= config["verbosity"]):
			for msg in entries:
				print(colorprefix + "[" + module + "] " + prefix + msg + colorsuffix)

		# file output
		if config["logfolder"] is not None:
			logfilename = config["logfolder"] + "/" + module + ".log"
			#os.makedirs(os.path.dirname(logfilename), exist_ok=True)
			with gopen(logfilename,"a") as logfile:
				for msg in entries:
					logfile.write(now + "  " + prefix + msg + "\n")
			trim(logfilename)


def flush():
	"""Outputs and empties the complete backlog."""
	global _queue
	for entry in _queue:
		# console output
		if entry["console"]:
			print("[" + entry["module"] + "] " + entry["prefix"] + entry["msg"])

		# file output
		if config["logfolder"] is not None:
			logfilename = config["logfolder"] + "/" + entry["module"] + ".log"
			#os.makedirs(os.path.dirname(logfilename), exist_ok=True)
			with gopen(logfilename,"a") as logfile:
				logfile.write(entry["time"] + "  " + entry["prefix"] + entry["msg"] + "\n")
			trim(logfilename)

	_queue = []

def trim(filename):

	try:
		#sanity check for old version logfiles
		if os.path.getsize(filename) > config["maxsize"] * 4:
			os.remove(filename)

		elif os.path.getsize(filename) > config["maxsize"]:
			with open(filename,"r") as sourcefile:
				lines = sourcefile.readlines()
				delete = int(len(lines)/2)
			with open(filename,"w") as targetfile:
				targetfile.writelines(lines[delete:])
	except Exception:
		pass


# Quicker way to add header
def logh1(*args,**kwargs):
	"""Logs a top-level header. Otherwise, same arguments as :func:`log`"""
	return log(*args,**kwargs,header=1)
def logh2(*args,**kwargs):
	"""Logs a second-level header. Otherwise, same arguments as :func:`log`"""
	return log(*args,**kwargs,header=2)



from threading import Thread, Timer
import datetime
import random
import time

from ._internal import DEFAULT, defaultarguments, gopen, DoreahConfig


config = DoreahConfig("regular",
	autostart=True,
	offset=0
)

def tz():
	return datetime.timezone(offset=datetime.timedelta(hours=config["offset"]))

### DEPRECATED
def yearly(func):
	"""Decorator for yearly functions. Depending on configuration, is equivalent to either :meth:`runyearly` or :meth:`repeatyearly`."""
	if config["autostart"]: return runyearly(func)
	else: return repeatyearly(func)

def monthly(func):
	"""Decorator for monthly functions. Depending on configuration, is equivalent to either :meth:`runmonthly` or :meth:`repeatmonthly`."""
	if config["autostart"]: return runmonthly(func)
	else: return repeatmonthly(func)

def daily(func):
	"""Decorator for daily functions. Depending on configuration, is equivalent to either :meth:`rundaily` or :meth:`repeatdaily`."""
	if config["autostart"]: return rundaily(func)
	else: return repeatdaily(func)

def hourly(func):
	"""Decorator for hourly functions. Depending on configuration, is equivalent to either :meth:`runhourly` or :meth:`repeathourly`."""
	if config["autostart"]: return runhourly(func)
	else: return repeathourly(func)
###


intervals = {
	'yearly':{
		'getnext':lambda now:datetime.datetime(now.year+1,1,1,tzinfo=tz()),
		'functions':[]
	},
	'monthly':{
		'getnext':lambda now:datetime.datetime(now.year,now.month + 1,1,tzinfo=tz()) if now.month != 12 else datetime.datetime(now.year+1,1,1,tzinfo=tz()),
		'functions':[]
	},
	'daily':{
		'getnext':lambda now:datetime.datetime(now.year,now.month,now.day,tzinfo=tz()) + datetime.timedelta(days=1),
		'functions':[]
	},
	'hourly':{
		'getnext':lambda now:datetime.datetime(now.year,now.month,now.day,now.hour,tzinfo=tz()) + datetime.timedelta(hours=1),
		'functions':[]
	},
	# just for testing
	'constantly':{
		'getnext':lambda now:datetime.datetime.now(tz=tz()) + datetime.timedelta(seconds=1),
		'functions':[]
	}
}

# init daemon for each interval
# spawns all runs, then goes to sleep until next time
def doreah_regular_daemon(interval):
	i = 0
	while True:
		i += 1
		now = datetime.datetime.now(tz=tz())
		next = intervals[interval]['getnext'](now)

		wait = int(next.timestamp() - now.timestamp()) + 5

		time.sleep(wait)
		for f in intervals[interval]['functions']:
			try:
				t = Thread(target=f['function'],args=f['args'],kwargs=f['kwargs'])
				t.daemon = False
				t.setName(f"doreah-regular-{interval}-loop{i}-{f['function'].__name__}")
				t.start()
				#f['function'](*f['args'],**f['kwargs'])
			except Exception as e:
				print(e)
			time.sleep(2)

# spawns above init daemons if any functions actually need it
def spawn_all_needed():

	for interval in intervals:
		data = intervals[interval]
		if len(data['functions']) > 0 and not data.get('thread'):
			data['thread'] = Thread(target=doreah_regular_daemon,kwargs={'interval':interval})
			data['thread'].daemon = True
			data['thread'].setName(f"doreah-regular-{interval}-init")
			data['thread'].start()


def run_regularly(interval,func):

	functioninfo = {
		'function':func,
		'args':(),
		'kwargs':{}
	}

	# run it once
	t = Timer(5,**functioninfo)
	t.daemon = True
	t.start()

	# add to list
	intervals[interval]['functions'].append(functioninfo)
	spawn_all_needed()

	# return unchanged function
	return func

def repeat_regularly(interval,func):

	def self_scheduling_func(*args,**kwargs):

		functioninfo = {
			'function':func,
			'args':args,
			'kwargs':kwargs
		}

		# execute function
		func(*args,**kwargs)

		# add original function to list, with args and kwargs from first invocation
		intervals[interval]['functions'].append(functioninfo)
		spawn_all_needed()

	return self_scheduling_func


def runyearly(func):
	"""Decorator to make the function execute on first definition as well as every year."""
	return run_regularly('yearly',func)


def repeatyearly(func):
	"""Decorator to make the function repeat every new year after being called once."""
	return repeat_regularly('yearly',func)

def runmonthly(func):
	"""Decorator to make the function execute on first definition as well as every month."""
	return run_regularly('monthly',func)

def repeatmonthly(func):
	"""Decorator to make the function repeat every new month after being called once."""
	return repeat_regularly('monthly',func)

def rundaily(func):
	"""Decorator to make the function execute on first definition as well as every day."""
	return run_regularly('daily',func)

def repeatdaily(func):
	"""Decorator to make the function repeat every new day after being called once."""
	return repeat_regularly('daily',func)

def runhourly(func):
	"""Decorator to make the function execute on first definition as well as every hour."""
	return run_regularly('hourly',func)

def repeathourly(func):
	"""Decorator to make the function repeat every new hour after being called once."""
	return repeat_regularly('hourly',func)

# just for testing
def runoften(func):
	return run_regularly('constantly',func)



from ._internal import DoreahConfig

import yaml
import random
from jinja2 import Environment, PackageLoader, select_autoescape


config = DoreahConfig("keystore",
	key_length=64
)

JINJAENV = Environment(
	loader=PackageLoader('doreah', 'resources/keystore'),
	autoescape=select_autoescape(['html', 'xml'])
)



charset = list(range(10)) + list("abcdefghijklmnopqrstuvwxyz") + list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")
def randomstring(length):
	return "".join(str(random.choice(charset)) for _ in range(length))


class KeyStore:
	"""Object that represents keys and their identifiers stored by a user. They are accessible via a web interface
	and written to a permanent file."""

	def __init__(self,file="keystore.yml",save_endpoint="/keys"):
		self.file = file
		self.keys = {}
		self.save_endpoint = save_endpoint
		self.load_from_file()

	def __len__(self):
		return self.keys.__len__()
	def __getitem__(self,identifier):
		return self.keys[identifier]
	def __setitem__(self,identifier,value):
		self.keys[identifier] = value
		self.write_to_file()

	def __iter__(self):
		return (identifier for identifier in self.keys)

	def check_key(self,key):
		return (key in self.keys.values())

	def check_and_identify_key(self,key):
		for identifier in self.keys:
			if self.keys[identifier] == key:
				return identifier
		return False

	def generate_key(self,identifier):
		key = randomstring(config['key_length'])
		self.keys[identifier] = key
		self.write_to_file()
		return key

	def add_key(self,identifier,key):
		self.keys[identifier] = key
		self.write_to_file()

	def delete_key(self,identifier):
		del self.keys[identifier]
		self.write_to_file()

	def update(self,dict):
		for identifier in dict:
			if dict[identifier] is None:
				self.delete_key(identifier)
			else:
				self.add_key(identifier,dict[identifier])

	def load_from_file(self):
		try:
			with open(self.file) as filed:
				self.keys = yaml.safe_load(filed)
		except FileNotFoundError:
			self.write_to_file()

	def write_to_file(self):
		with open(self.file,"w") as filed:
			yaml.dump(self.keys,filed)

	def html(self):
		template = JINJAENV.get_template("settings.jinja")
		return template.render({"keystore":self})
